// SPDX-License-Identifier: GPL-2.0-only

/*
 *  ZeroMAC console library
 *
 *  Author:
 *      Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *  Copyright (C) 2015-2020 Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2,
 *      as published by the Free Software Foundation.
 */
 
#include "libzmac_console.h"


//
// zmac_allocate_command_and_args
//
// It is a helper for the ZeroMAC UI user space utilities, it will pack the string you give with all the args from the shell
// separated by space into one string, which is allocated by malloc, so you have to free it after use.
// It will also set the length of the allocated memory to an integer if you set it's address in paramter.
// If there are no args, and you do not give a string to start with, it will malloc 0 size and will set 0 as the length 
// of the allocated block (and will not write closing 0), so you have to care about the length of the block, or the 
// number of arguments. Arguments will be handled only from the second one, as the first is the name of the application you 
// run currently. The allocated length of course counted with the string closing '\0'.
// If there is an LF in the arg strings, it will cut at LF, and the shrinked string and length returned.
//
// command	- address of 0 terminated string as a prefix in the allocated string, or NULL if you don't have prefix
// argc		- the number of arguments you would like to add
// argv		- pointer to the start of the arguments (pointer) array
// len		- pointer to the integer where to store the allocated length
//
// RETURN VALUE	- the address of the allocated string
//


LIB_ZMAC_CONSOLE_NEED_INLINE char *zmac_allocate_command_and_args(char *command,int argc,char **argv,int *len)
{
	register int l,c;
	char *b,*w,*a=argv[1];

	for (l=argc+(command?strlen(command):-1),c=1;c<argc;c++) l+=strlen(argv[c]);			// closing 0 included in length, or 0 length if no content at all
	if (!(w=b=malloc(l))) { fprintf(stderr,ZMAC_LOG "Out of memory error!\n"); return NULL;}
	if (command) for (;(*w=*command++);w++);
	for (c=1;c<argc;a=argv[++c]) 
	{
		if (w!=b) *w++=' ';
		while ((*w=*a++)) 
		{
			if (*w=='\n') {*w=0; c=argc; l=w-b+1; break;}					// LF not allowed here, that could be an attack against userspace utility
			w++;
		}
	} 
	if (len) *len=l;
	return b;
}


//
// zmac_api_parse_response     
//
// It will parse a console response, will return the length, so you can call it in a row to get answers. 
// It can understand and handle multiline responses too, but do not suggest to use it, because if you are under attack, 
// and attacker can flood your labels and rules, the answer can be much larger than you expect it.
// 
// 	              
// buffer       - 0 terminated api block from the ZeroMAC console, string will be modified, and cannot be used after parsing the response.
// result_str   - the address of the pointer which will point to the result string
// content      - the address of the pointer which will point to the content string
// length       - the address of the int variable which will hold the size of the current processed block
//		  (you have to add this int to the buffer pointer to process the next block, if you have more)
//
// RETURN value - the ZeroMAC api response code, negative means errors from console, positive for result codes,
//		  0 means processing error, this case none of the return values can be used!
//


int zmac_api_parse_response(char *buffer,char **result_str,char **content,int *length)
{
	int r=0,negative=0;
	char *p=buffer;

	*content=NULL;
	if (*p!='[') return 0;							// valid result must start with 'A'nswer ('A'cknowledge) or 'E'rror, otherwise this is an invalid API result
	p++;											
	if (*p=='-')								// negative numbers are error codes
	{
		negative=1;
		p++;
	}
	while (*p>='0' && *p<='9') r=r*10+*p++-'0';				// result code is a decimal number
	if (negative) r=-r;
	if (!*p) return 0;							
	if (!r) return 0;							// this is invalid, if there is no number
	if (*p!=']')								// if there is no closing ], this is a multiline result
	{
		while (*p && *p!='\n') p++;
		if (!*p) return 0;
		*content=++p;							// content starts from the beginning of the next line
		while (*p)							// let's see the lines
		{
			if (*p=='[')
			{
				*p=0;						// this is the end of the content
				p++;
				negative=r=0;
				if (*p=='-')					// negative numbers are error codes
				{	
					negative=1;
					p++;
				}
				while (*p>='0' && *p<='9') r=r*10+*p++-'0';	// result code is a decimal number
				if (negative) r=-r;
				if (!*p) return 0;				
				if (!r) return 0;				// this is invalid, if there is no number
				if (*p!=']') return 0;
				while (*p==' ') p++;
				if (!*p) return 0;							
				*result_str=p;
				while (*p && *p!='\n') p++;
				if (!*p) return 0;
				*p=0;						// the end of result string
				*length=p-buffer+1;				// set the length of the block 
				return r;					// return multiline
			}
			else
			{
				while (*p && *p!='\n') p++;
				if (!*p) return 0;
				p++;
			}
		}
		return 0;							// if here is the end of the string, then this is an invalid multiline block
	}
	p++;
	while (*p==' ') p++;							// right after the API result code there is a space
	*result_str=p;								// here is the beginning of the result string
	while (*p && *p!=':' && *p!='\n') p++;					// have to find it after ':'
	if (*p==':') 
	{
		p++;
		while (*p==' ') p++;
		*content=p;
		if (!*p) return 0;
	}									// if there is content, set it (after a space if we have one)
	else *content=NULL;
	while (*p && *p!='\n') p++;						// find the end of the line
	if (!*p) return 0; 							// if there is no closing enter, the block is invalid
	*p=0;									// change LF to closing 0 (this is the end of the result string and the single line content if we have)
	*length=p-buffer+1;							// set the length of the block 
	return r;								// on error return result code as negative, on success, return it as positive
}



//
// zmac_open_console
//
// It will open the ZeroMAC user interface console, and return a file descriptor.
// 
// RETURN VALUE - non negativ value is the file descriptor of the opened console, 
//		  negative value means error (which is a ZeroMAC utility error return code)
//		  
//



static int zmac_open_console()
{
	char *console_fn=ZMAC_CONSOLE;
	int fd;
	//if ((env_fn=getenv("ZMAC_CONSOLE"))) console_fn=env_fn;
	if ((fd=open(console_fn,O_RDWR))<0) 
	{
		fprintf(stderr,ZMAC_LOG "Error: cannot open ZeroMAC console. (%s)\n",console_fn);
		return ZMAC_ERROR_CANNOT_OPEN_CONSOLE;
	}
	return fd;
}



//
// zmac_console_submit
//
// Will submit commands to the ZeroMAC console, and receive results. Raw console answer will be 
// stored in input buffer.
//
// output_buffer	- 0 terminated string, output command(s) separated by '\n'
// output_buffer_size	- the size of the commands, you should include the closing 0, that means EOF for the console
// input_buffer		- the allocated input buffer to read results into
// input_buffer_size	- the total size of the input buffer
//
// RETURN VALUE	- on succesful communication the size of the received messages, on error the negative 
//		  error code will be returned.
//		  ENOBUFS means you do not have enough space to store the results.
//		  On error you have to discard input buffer.
//


int zmac_console_submit(char *output_buffer,int output_buffer_size,char *input_buffer,int input_buffer_size)
{
	int fd;
	struct pollfd pfd;
	register int r,w=0,p=0;

	if ((fd=zmac_open_console())<0) return fd;

	for (;;)
	{
		pfd.fd=fd;
		pfd.events=POLLIN|(output_buffer_size-w?POLLOUT:0);
		pfd.revents=0;
		r=poll(&pfd,(unsigned int)1,ZMAC_CONSOLE_TIMEOUT);
		if (r<0) { if (errno==EINTR) continue; r=ZMAC_ERROR_POLL_FAILED; fprintf(stderr,ZMAC_LOG "Error: Poll failed!\n"); break;}
		if (r==0) { fprintf(stderr,ZMAC_LOG "Error: Timeout waiting for ZeroMAC console!\n"); r=ZMAC_ERROR_TIMEOUT; break;}
		if (pfd.revents & (POLLIN|POLLHUP))
		{
			r=read(fd,input_buffer+p,input_buffer_size-p);
			if (r>0) { p+=r; if (p==input_buffer_size) { fprintf(stderr,ZMAC_LOG "Error: Input buffer is too small!\n"); r=ZMAC_ERROR_INPUT_BUFFER_TOO_SMALL; break; } }
			else if (r<0) { fprintf(stderr,ZMAC_LOG "Error reading from ZeroMAC console!\n"); r=ZMAC_ERROR_CONSOLE_IO_ERROR; break; }
			else { r=p; break;}																																	
		}
		if ((pfd.revents & POLLOUT) && (output_buffer_size-w))
		{
			r=write(fd,output_buffer+w,output_buffer_size-w);
			if (r>=0) w+=r;
			else { fprintf(stderr,ZMAC_LOG "Error writing to ZeroMAC console!\n"); r=ZMAC_ERROR_CONSOLE_IO_ERROR; break; } 
		}
	}
	close(fd);
	return r;
}


//
// zmac_console_pipe
//
// Will pipe input and output file descriptors through the ZeroMAC console.
// Commands sent on input will be processed by the ZeroMAC UI console, and raw output
// of the console will be forwarded to the output file descriptor.
// You can open files it will read and write to them, or you can use STDIN and/or STDOUT to 
// use it as a real pipe
//
// fd_in	- input file descriptor
// fd_out	- output file desciptor
//
// RETURN VALUE	- 0 on success, or error code
// 		  will not return on some errors intentionally (because some errors could be normal, 
//		  and we should keep feeding atfer them) Errors are reported to stderr
//


int zmac_console_pipe(int fd_in,int fd_out)
{
	// fd_in read -> inbuff_write   inbuff_read -> fd_console write 
	// fd_console read -> outbuff_write outbuff_read -> fd_out write
	#define inbuff_free (ZMAC_CONSOLE_PIPE_BUFFER_SIZE-inbuff_write)
	#define outbuff_free (ZMAC_CONSOLE_PIPE_BUFFER_SIZE-outbuff_write)
	#define inbuff_size (inbuff_write-inbuff_read)
	#define outbuff_size (outbuff_write-outbuff_read)
	char inbuff[ZMAC_CONSOLE_PIPE_BUFFER_SIZE+1],outbuff[ZMAC_CONSOLE_PIPE_BUFFER_SIZE];
	struct pollfd pfd[3];
	register int r=0,inbuff_read=0,inbuff_write=0,outbuff_read=0,outbuff_write=0;
	register unsigned int pfdco;
	int eof_console=0,can_write_out=1,can_write_console=1,eof_in=0,poll_console,poll_in,poll_out,fd_console,rc_n=0,rc_negative=0,rc_status=1,rc=ZMAC_ERROR_NO_API_COMMAND_EXECUTED;

	if ((fd_console=zmac_open_console())<0) return fd_console;

	while ((!eof_console && can_write_console) || (outbuff_size && can_write_out))
	{
		if (inbuff_read && inbuff_read==inbuff_write) inbuff_read=inbuff_write=0;
		if (outbuff_read && outbuff_read==outbuff_write) outbuff_read=outbuff_write=0;
		if (inbuff_read && !inbuff_free) { memmove(inbuff,inbuff+inbuff_read,inbuff_size); inbuff_write=inbuff_size;inbuff_read=0; }
		if (outbuff_read && !outbuff_free) { memmove(outbuff,outbuff+outbuff_read,outbuff_size); outbuff_write=outbuff_size;outbuff_read=0; }

		pfdco=0;
		if ((poll_console=!eof_console && ((inbuff_size && can_write_console) || (outbuff_free))))
		{
			pfd[pfdco].fd=fd_console;
			pfd[pfdco].events=(inbuff_size && can_write_console?POLLOUT:0)+(outbuff_free && !eof_console?POLLIN:0);
			pfd[pfdco].revents=0;
			pfdco++;
		}
		if ((poll_in=inbuff_free&& !eof_in)) // inbuff_free && !eof_in
		{
			pfd[pfdco].fd=fd_in;
			pfd[pfdco].events=(inbuff_free?POLLIN:0) | POLLHUP;
			pfd[pfdco].revents=0;
			pfdco++;
		}
		if ((poll_out=can_write_out)) //outbuff_size && can_write_out
		{
			pfd[pfdco].fd=fd_out;
			pfd[pfdco].events=(outbuff_size?POLLOUT:0); // HUP-ot megkapjuk akkor is, ha nem akarunk adatot irni
			pfd[pfdco].revents=0;
			pfdco++;
		}
		r=poll(pfd,pfdco,ZMAC_CONSOLE_NO_TIMEOUT);
		if (r<0) { if (errno==EINTR) continue; fprintf(stderr,ZMAC_LOG "Error: Poll failed!\n"); rc=ZMAC_ERROR_POLL_FAILED; break;}
		if (r==0) { fprintf(stderr,ZMAC_LOG "Error: Timeout waiting for ZeroMAC console!\n"); rc=ZMAC_ERROR_TIMEOUT; break;} // this should not happen if there is no_timeout
		pfdco=0;
		if (poll_console)
		{
			if (outbuff_free && (pfd[pfdco].revents & (POLLIN|POLLHUP)))
			{
				r=read(fd_console,outbuff+outbuff_write,outbuff_free);
				if (r>0)
				{
					register int c=0;
					while (c<r)
					{
						switch (rc_status)
						{
							case 0:
								while (c<r) if (outbuff[outbuff_write+c]=='\n') break; else c++;
								if (c==r) break;
								else if (++c==r) { rc_status=1; break;}
							case 1:
								if (outbuff[outbuff_write+c]!='[') { rc_status=0; break;}
								if (++c==r) { rc_status=2; break;}
							case 2:
								if ((rc_negative=outbuff[outbuff_write+c]=='-') && ++c==r) { rc_status=3; break;}
							case 3:
								while (c<r && outbuff[outbuff_write+c]>='0' && outbuff[outbuff_write+c]<='9') rc_n=rc_n*10+outbuff[outbuff_write+c++]-'0';
								if (c==r) { rc_status=3; break;}
							case 4:
								if (outbuff[outbuff_write+c]==']') rc=rc_negative?-rc_n:rc_n;
								rc_n=rc_status=0;
						}
					}
					outbuff_write+=r;
				}
				else 
				{
					if (r<0) fprintf(stderr,ZMAC_LOG "Error reading from ZeroMAC console!\n");
					eof_console=1;
				}
			}
			if (pfd[pfdco].revents & POLLOUT)
			{
				r=write(fd_console,inbuff+inbuff_read,inbuff_size);
				if (r>0) inbuff_read+=r;
				else 
				{
					fprintf(stderr,ZMAC_LOG "Error writing to ZeroMAC console!\n");
					can_write_console=0;
				}
			}
			pfdco++;
		}
		if (poll_in)
		{
			if (pfd[pfdco].revents)
			{
				r=read(fd_in,inbuff+inbuff_write,inbuff_free);
				if (r>0) inbuff_write+=r;
				else 
				{
					if (r<0) fprintf(stderr,ZMAC_LOG "Error reading from input!\n");
					if (!r) inbuff[inbuff_write++]=0; // eof for the zmac console
					eof_in=1;
				}
			}
			pfdco++;
		}
		if (poll_out)
		{
			if (pfd[pfdco].revents & POLLHUP) can_write_out=outbuff_read=outbuff_write=0;
			else if (pfd[pfdco].revents)
			{
				r=write(fd_out,outbuff+outbuff_read,outbuff_size);
				if (r>=0) outbuff_read+=r;
				else 
				{
					fprintf(stderr,ZMAC_LOG "Error writing to output!\n");
					can_write_out=outbuff_read=outbuff_write=0;
				}
			}
		}
	}
	close(fd_console);
	memset(inbuff,0,ZMAC_CONSOLE_PIPE_BUFFER_SIZE+1);
	memset(outbuff,0,ZMAC_CONSOLE_PIPE_BUFFER_SIZE);
	return rc;
}



//
// zmac_console_submit_and_print_result
//
// Submit a command to ZeroMAC console, and print the result to stdout.
// The raw console output will be printed!
// This is mostly for debug, will be removed....
//
// buff		- the command(s) to submit terminated with 0, commands are separated with '\n'
// size		- the length of the commands
// 
// RETURN VALUE	- 0 means success, otherwise error code returned
//		  Note, API response codes will not be handled and returned, so returns 0 if the command sent, 
//		  and answer received without communication errors with the console
//


int zmac_console_submit_and_print_result(char *buff,int size)
{
	char inbuff[ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE+1];
	struct pollfd pfd;
	register int r,w=0,fd;

  if ((fd=zmac_open_console())<0) return fd;
	
	for (;;)
	{
		pfd.fd=fd;
		pfd.events=POLLIN|(size-w?POLLOUT:0);
		pfd.revents=0;
		r=poll(&pfd,(unsigned int)1,ZMAC_CONSOLE_TIMEOUT);
		if (r<0) { if (errno==EINTR) continue; fprintf(stderr,ZMAC_LOG "Error: Poll failed!\n"); r=ZMAC_ERROR_POLL_FAILED; break;}
		if (r==0) { fprintf(stderr,ZMAC_LOG "Error: Timeout waiting for ZeroMAC console!\n"); r=ZMAC_ERROR_TIMEOUT; break;}
		if (pfd.revents & (POLLIN|POLLHUP))
		{
			r=read(fd,inbuff,ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE);
			if (r>0) 
			{
				inbuff[r]=0;
				printf("%s",inbuff);
			}
			else if (r<0) { fprintf(stderr,ZMAC_LOG "Error reading from ZeroMAC console!\n"); r=ZMAC_ERROR_CONSOLE_IO_ERROR; break; }
			else break;																							
		}
		if ((pfd.revents & POLLOUT) && (size-w))
		{
			r=write(fd,buff+w,size-w);
			if (r>0) w+=r;
			else { fprintf(stderr,ZMAC_LOG "Error writing to ZeroMAC console!\n"); r=ZMAC_ERROR_CONSOLE_IO_ERROR; break; } 
		}
	}
	close(fd);
	memset(inbuff,0,ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE+1);
	return r;
}





//
// zmac_console_submit_and_print_result_api
//
// Submit a command to ZeroMAC console, and print the result to stdout.
// It is not memory buffer size limited, it will print the content of the zeromac console answers,
// and will get and return the ZeroMAC api status code. 
//
// buff		- the command(s) to submit terminated with 0, commands are separated with '\n'
// size		- the length of the commands
// fdout	- the device number of the output device ( 1 for stdout)
// fderr	- the device number of the error device ( 2 for stderr, 0 will hide error messages)
// print_status_only_on_error - 1 means will print status messages only if error, 0 - means all status messages will be printed
// 
// RETURN VALUE	- 0 means success, otherwise error code returned
//		  Note, API response codes will be returned 
//		  
//





int zmac_console_submit_and_print_result_api(char *buff,int size,int fdout,int fderr,int print_status_only_on_error)
{
	char inbuff[ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE+1];
	struct pollfd pfd;
	register int c,r,n,w=0,api_status=0,multiline=0;
	int fd,response_code=ZMAC_ERROR_NO_API_COMMAND_EXECUTED,printbuff,newline,content,negative=0;

	if ((fd=zmac_open_console())<0) return fd;
		
	for (;;)
	{
		pfd.fd=fd;
		pfd.events=POLLIN|(size-w?POLLOUT:0);
		pfd.revents=0;
		r=poll(&pfd,(unsigned int)1,ZMAC_CONSOLE_TIMEOUT);
		if (r<0) { if (errno==EINTR) continue; response_code=ZMAC_ERROR_POLL_FAILED; fprintf(stderr,ZMAC_LOG "Error: Poll failed!\n"); break;}
		if (r==0) { fprintf(stderr,ZMAC_LOG "Error: Timeout waiting for ZeroMAC console!\n"); response_code=ZMAC_ERROR_TIMEOUT; break;}
		if (pfd.revents & (POLLIN|POLLHUP))
		{
			r=read(fd,inbuff,ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE);
			if (r>0) 
			{
				c=0;
				switch(api_status)
				{
				restart:
					case 0:
						if (c==r) { api_status=0; break;}
						if (inbuff[c]!='[') { content=1; goto print_content;}			// at the beginning of the line there must be a [, if not, then this is a content line
						negative=n=0;
						c++;
						if (c==r) { api_status=1; break;}
					case 1:
						if (inbuff[c]=='-') { negative=1; c++;}
						if (c==r) { api_status=2; break;}
					case 2:						
						while (c<r && inbuff[c]>='0' && inbuff[c]<='9') n=n*10+inbuff[c++]-'0';	// after [ there must be a numeric code
						if (negative) n=-n;
						if (c==r) { api_status=3; break;}
					case 3:
						if (inbuff[c]==']')							// if there is a closing ]
						{
							response_code=n;						// then this code is the response code
							content=!multiline;						// if this is a multiline block, then this is the status, which is important only if negative, if it is not a multiline block, then this is the content itself, we have to display to std output
							multiline=0;
							c++;
							if (c==r) { api_status=4; break;}
						}
						else									// if not closed with ], then this is the first line of a multiline response, or one of the inline status lines in a multi line response
						{
							multiline=1;
							if (n>0) goto discard_content;					// if the code is positive, then this is the title of a list, or acknowledge of batched commands, transactions, discard the rest of the line
							content=0;							// if the code is negative, and this is not the end of the block, then this is an error text, print to std error
						}
					case 4:
						while (c<r && inbuff[c]==' ') c++;					// have some white space
						if (c==r) { api_status=4; break;}
				print_content:
					case 5:
						printbuff=c;
						while (c<r && inbuff[c]!='\n') c++;
						if (c!=r) { c++; newline=1;}
						if (content || response_code<0 || !print_status_only_on_error)
						{
							int w;
						write_next_turn:
        						w=write(negative?fderr:fdout,inbuff+printbuff,c-printbuff);
        						if (w>0)
        						{
        							printbuff+=w;
        							if (printbuff!=c) goto write_next_turn;
        						}
        						else if (w<0)
        						{
        							if (w==-EAGAIN) goto write_next_turn;
        							response_code=ZMAC_ERROR_CONSOLE_IO_ERROR;
								goto cleanup;
        						}
						}
						if (newline) goto restart;
						api_status=5;
						break;
				discard_content:
					case 6:
						while (c<r && inbuff[c]!='\n') c++;
						if (c==r) { api_status=6; break;}
						c++;
						goto restart;
				}
			}
			else if (r<0) { response_code=ZMAC_ERROR_CONSOLE_IO_ERROR; fprintf(stderr,ZMAC_LOG "Error reading from ZeroMAC console!\n"); break; }
			else break;																							
		}
		if ((pfd.revents & POLLOUT) && (size-w))
		{
			r=write(fd,buff+w,size-w);
			if (r>0) w+=r;
			else { response_code=ZMAC_ERROR_CONSOLE_IO_ERROR; fprintf(stderr,ZMAC_LOG "Error writing to ZeroMAC console!\n"); break; } 
		}
	}
cleanup:
	close(fd);
	memset(inbuff,0,ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE+1);
	return response_code;
}





// 
// zmac_run_console_command
// 
// Submit a command to ZeroMAC console, and returns with shell exit code.
// The command(s) must be started with API req, to have API response codes in reply.
// The first API response code will be handled, you can do more silently.
// Remember, ZeroMAC console is async, it will run all the commands you submit (until output buffer is full) 
// but you will see the result only the first which generates answer!
// Also note, it was designed to use simple commands, works only with answers smaller than the receive buffer.
// (Do not use commands like 'show rules', because those can reply much longer than the buffer,
// use another method for commands with variable length answers)
//
// command	- 0 terminated command string to submit
// len		- the length of the command(s), you should include the closing 0 in length, that means EOF for the console
// print	- what part to print (ZMAC_PRINT_RESULT or ZMAC_PRINT_CONTENT or 0)
// 
// RETURN	- shell exit code, 0 means ok, other than 0 means error
//		  API response codes will be handled and returned on error, 0 will be returned on positive API response codes 
//

LIB_ZMAC_CONSOLE_NEED_INLINE int zmac_run_console_command(char *command,int len,int print)
{
	register int r;
	char result[ZMAC_RUN_CONSOLE_COMMAND_BUFFER_SIZE],*result_str,*content;
	if ((r=zmac_console_submit(command,len,result,ZMAC_RUN_CONSOLE_COMMAND_BUFFER_SIZE-1))<0) return r;
	result[r]=0;
	r=zmac_api_parse_response(result,&result_str,&content,&len);
	if (r>0)
	{
		if (print==ZMAC_PRINT_RESULT) printf("%s\n",result_str); 
		else if (print==ZMAC_PRINT_CONTENT) printf("%s\n",content?content:"");
		r=0; 
	}
	else if (!r)  { printf(ZMAC_LOG "Error: Invalid api response from ZeroMAC console!\n"); r=ZMAC_ERROR_INVALID_API_RESPONSE; }
	else if (r<0) { printf(ZMAC_LOG "Error: %s\n",result_str); }
	memset(result,0,ZMAC_RUN_CONSOLE_COMMAND_BUFFER_SIZE);
	return r;
}

//
// zmac_return_shell_code
//
// ZeroMAC error codes are negative numbers, but we have to return them as positive to the shell, and 0
// if there are positive ZeroMAC return codes.
//

inline int zmac_return_shell_code(int rc)
{
	return rc<0?-rc:0;
}