// SPDX-License-Identifier: GPL-2.0-only

/*
 *  ZeroMAC console library
 *
 *  Author:
 *      Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *  Copyright (C) 2015-2020 Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2,
 *      as published by the Free Software Foundation.
 */



#ifndef LIBZMAC_CONSOLE_H
#define LIBZMAC_CONSOLE_H

#include <stdio.h>  
#include <unistd.h>  
#include <stdlib.h>  
#include <string.h>  
#include <sys/types.h>
#include <sys/xattr.h>
#include <errno.h>
#include <poll.h>
#include <fcntl.h>

#define ZMAC_LOG "ZMAC: "
#ifndef ZMAC_CONSOLE
#define ZMAC_CONSOLE "/dev/zmac"
#endif

#define SECONDS *1000
#define ZMAC_CONSOLE_TIMEOUT 60 SECONDS
#define ZMAC_CONSOLE_NO_TIMEOUT -1

#define ZMAC_CONSOLE_SUBMIT_INPUT_BUFFER_SIZE 4096
#define ZMAC_CONSOLE_PIPE_BUFFER_SIZE 4096
#define ZMAC_RUN_CONSOLE_COMMAND_BUFFER_SIZE 4096

#define ZMAC_PRINT_RESULT 1
#define ZMAC_PRINT_CONTENT 2

// error code 1-2 reserved for the shell, 3-19 reserved for the user space utilities, 20-99 are reserved for kernel UI errors
#define ZMAC_ERROR_CANNOT_OPEN_CONSOLE 		-3
#define ZMAC_ERROR_TIMEOUT 			-4
#define ZMAC_ERROR_CONSOLE_IO_ERROR 		-5
#define ZMAC_ERROR_POLL_FAILED 			-6
#define ZMAC_ERROR_INVALID_API_RESPONSE 	-7
#define ZMAC_ERROR_OUT_OF_MEMORY 		-8
#define ZMAC_ERROR_INPUT_BUFFER_TOO_SMALL 	-9
#define ZMAC_ERROR_INVALID_LABEL		-10
#define ZMAC_ERROR_NO_API_COMMAND_EXECUTED	-11



//
// zmac_allocate_command_and_args
//
// It is a helper for the ZeroMAC UI user space utilities, it will pack the string you give with all the args from the shell
// separated by space into one string, which is allocated by malloc, so you have to free it after use.
// It will also set the length of the allocated memory to an integer if you set it's address in paramter.
// If there are no args, and you do not give a string to start with, it will malloc 0 size and will set 0 as the length 
// of the allocated block (and will not write closing 0), so you have to care about the length of the block, or the 
// number of arguments. Arguments will be handled only from the second one, as the first is the name of the application you 
// run currently. The allocated length of course counted with the string closing '\0'.
// If there is an LF in the arg strings, it will cut at LF, and the shrinked string and length returned.
//
// command	- address of 0 terminated string as a prefix in the allocated string, or NULL if you don't have prefix
// argc		- the number of arguments you would like to add
// argv		- pointer to the start of the arguments (pointer) array
// len		- pointer to the integer where to store the allocated length
//
// RETURN VALUE	- the address of the allocated string
//


// static inline char *zmac_allocate_command_and_args(char *command,int argc,char **argv,int *len);


//
// zmac_api_parse_response     
// 	              
// buffer       - 0 terminated api block from the ZeroMAC console, string will be modified, and cannot be used after parsing the response.
// result_str   - the address of the pointer which will point to the result string
// content      - the address of the pointer which will point to the content string
// length       - the address of the int variable which will hold the size of the current processed block
//		  (you have to add this int to the buffer pointer to process the next block, if you have more)
//
// RETURN value - the ZeroMAC api response code, negative means errors from console, positive for result codes,
//		  0 means processing error, this case none of the return values can be used!
//


int zmac_api_parse_response(char *buffer,char **result_str,char **content,int *length);

//
// zmac_console_submit
//
// Will submit commands to the ZeroMAC console, and receive results. Raw console answer will be 
// stored in input buffer.
//
// output_buffer	- 0 terminated string, output command(s) separated by '\n'
// output_buffer_size	- the size of the commands, you should include the closing 0, that means EOF for the console
// input_buffer		- the allocated input buffer to read results into
// input_buffer_size	- the total size of the input buffer
//
// RETURN VALUE	- on succesful communication the size of the received messages, on error the negative 
//		  error code will be returned.
//		  ENOBUFS means you do not have enough space to store the results.
//		  On error you have to discard input buffer.
//


int zmac_console_submit(char *output_buffer,int output_buffer_size,char *input_buffer,int input_buffer_size);

//
// zmac_console_pipe
//
// Will pipe input and output file descriptors through the ZeroMAC console.
// Commands sent on input will be processed by the ZeroMAC UI console, and raw output
// of the console will be forwarded to the output file descriptor.
// You can open files it will read and write to them, or you can use STDIN and/or STDOUT to 
// use it as a real pipe
//
// fd_in	- input file descriptor
// fd_out	- output file desciptor
//
// RETURN VALUE	- 0 on success, or error code
//


int zmac_console_pipe(int fd_in,int fd_out);

//
// zmac_console_submit_and_print_result
//
// Submit a command to ZeroMAC console, and print the result to stdout.
// The raw console output will be printed!
//
// buff		- the command(s) to submit terminated with 0, commands are separated with '\n'
// size		- the length of the commands
// 
// RETURN VALUE	- 0 means success, otherwise error code returned
//		  Note, API response codes will not be handled and returned, so returns 0 if the command sent, 
//		  and answer received without communication errors with the console
//

int zmac_console_submit_and_print_result(char *buff,int size);

// 
// zmac_run_console_command
// 
// Submit a command to ZeroMAC console, and returns with shell exit code.
// The command(s) must be started with API req, to have API response codes in reply.
// The first API response code will be handled, you can do more silently.
// Remember, ZeroMAC console is async, it will run all the commands you submit (until output buffer is full) 
// but you will see the result only the first which generates answer!
// Also note, it was designed to use simple commands, works only with answers smaller than the receive buffer.
// (Do not use commands like 'show rules', because those can reply much longer than the buffer,
// use another method for commands with variable length answers)
//
// command	- 0 terminated command string to submit
// len		- the length of the command(s), you should include the closing 0 in length, that means EOF for the console
// print	- what part to print (ZMAC_PRINT_RESULT or ZMAC_PRINT_CONTENT or 0)
// 
// RETURN	- shell exit code, 0 means ok, other than 0 means error
//		  API response codes will be handled and returned on error, 0 will be returned on positive API response codes 
//

inline int zmac_run_console_command(char *command,int len,int print);


//
// zmac_return_shell_code
//
// ZeroMAC error codes are negative numbers, but we have to return them as positive to the shell, and 0
// if there are positive ZeroMAC return codes.
//

inline int zmac_return_shell_code(int rc);



#ifdef LIB_ZMAC_CONSOLE_INLINE
#define LIB_ZMAC_CONSOLE_NEED_INLINE static inline // if compiled with lib included, you can choose to build inline the needed ones
#else // LIB_ZMAC_CONSOLE_INLINE
#define LIB_ZMAC_CONSOLE_NEED_INLINE // if compiled as library, functions will appear
#endif // LIB_ZMAC_CONSOLE_INLINE


#endif // LIBZMAC_CONSOLE_H
