// SPDX-License-Identifier: GPL-2.0-only

/*
 *  ZeroMAC security label library
 *
 *  Author:
 *      Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *  Copyright (C) 2015-2020 Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2,
 *      as published by the Free Software Foundation.
 */


#include "libzmac_label.h"

LIB_ZMAC_LABEL_NEED_INLINE int zmac_getlabel(char *fn,char *label) // 0=olvasasi hiba, nincs foglalas 1=ok, foglalt blokk visszaadva
{
	register int l;
	l=lgetxattr(fn,"security.zmac",label,MAX_STORAGE_LABEL_LENGTH+1);
	if (!l || (l<0 && errno==ENODATA)) {*label=0; return 1;} // ha nincs label, vagy ures adat, de oke az olvasas
	if (l<0) { perror("Cannot read label from file. "); return 0;} // olvasasi hiba
	if (l>MAX_STORAGE_LABEL_LENGTH) { fprintf(stderr,"Storage label is too long.\n"); return 0;} // olvasasi hiba
	label[l]=0; // valid string, minden ok
	return l;
}

LIB_ZMAC_LABEL_NEED_INLINE int zmac_setlabel(const char *fn,const char *label,size_t size) // 0=hiba 1=ok, kell innen hiba? ===
{
	return !lsetxattr(fn,"security.zmac",label,size,0);
}

LIB_ZMAC_LABEL_NEED_INLINE int zmac_removelabel(const char *fn) // 0=error 1=ok
{
	return !lremovexattr(fn,"security.zmac");
}

int zmac_validlabel(char *s) // 0=nem 1=igen
{
	register int c;
	#define validlabelchar(ZZZ) ((ZZZ>='0' && ZZZ<='9')||(ZZZ>='a' && ZZZ<='z')||(ZZZ>='A' && ZZZ<='Z')||ZZZ=='+'||ZZZ=='-'||ZZZ=='_')
	if (!s) return 1; 						// ha ures, az valid!!! === hmmm itt most, de mint lib, ez nem oke
	if (strlen(s)>MAX_LABEL_LENGTH) return 0;
	for (c=0;s[c];c++) if (!validlabelchar(s[c])) return 0; 	// ha invalid char van, akkor hiba
	return 1;							// most a tuti
}







// ha valamelyik label nincs, akkor az NULL lesz
int zmac_extractmultiblock(char *s,size_t len,char **al,char **el,char **wl,char **otherblocks, int *otherblockslen) // visszaadja access labelt, exec labelt, kapja a stringet, aminek valid stringnek kell lennie, visszateres 0=invalid lebel 1=ok
{
	register char *label;
	register int p=2;
	if (!s || len<4 || !(s[0]=='L' && s[1]=='=')) return 0;	// invalid
	label=s+p;
	while (p<len && s[p] && s[p]!=':') p++;
	if (p==len) goto invalid_block;
	*al=label;
	if (!s[p]) goto endblock;
	s[p++]=0;
	label=s+p;
	while (p<len && s[p] && s[p]!=':') p++;
	if (p==len) goto invalid_block;
	*el=label;
	if (!s[p]) goto endblock;
	s[p++]=0;
	label=s+p;
	while (p<len && s[p]) p++;
	if (p==len) goto invalid_block;
	*wl=label;
endblock:
	p++;							// two closing zeros
	if (p<len && s[p]) p++;
	if (p==len) return 1;
	*otherblocks=s+p;					// the rest of the blocks, keep them
	*otherblockslen=len-p;					// until the end 
	return 1;						// valid
invalid_block:
	*al=*el=*wl=NULL;
	return 0;						// invalid
}



// use only with 0 closed strings
// modifies the input string
// no field validations here, returns the number of fields we have, 0 if none
int zmac_extractsingleblock(char *s,char **al,char **el,char **wl)
{
	register int p=0,v=0;
	while (s[p] && s[p]!=':') p++;				// first field is access label
	if (p) { v++;*al=s; }					// valid if it has chars
	if (!s[p]) return v;
	s[p++]=0;						// replace colon with zero
	s+=p;
	p=0;
	while (s[p] && s[p]!=':') p++;
	if (p) { v++; *el=s; }					// exec label 
	if (!s[p]) return v;
	s[p++]=0;						// replace colon with zero
	s+=p;
	p=0;
	while (s[p]) p++;
	if (p) { v++; *wl=s; }					// create label
	return v;
}


int zmac_composelabels(char *createlabel,char *a,char *e,char *c,char *otherblocks,int otherblockslen)
{
	register int lena=0,lene=0,lenc=0,writelen=0,size;
	if (a) lena=strlen(a);					// ha van access label
	if (e) lene=strlen(e);					// ha van exec label
	if (c) lenc=strlen(c);					// ha van create label
	size=lena+lene+lenc+otherblockslen+(lenc?2:(lene?1:0))+(otherblockslen?4:0);
	if (!size) return 0;		// delete label completely
	if (size>MAX_STORAGE_LABEL_LENGTH) return ZMAC_ERROR_SECLABEL_SIZE_EXCEEDED;
	if (otherblockslen) { createlabel[writelen++]='L'; createlabel[writelen++]='=';}	// if we have other blocks to write we have to head this block, from 9.5 drop singleblock, must head!!! ===
	if (lena) { memcpy(createlabel+writelen,a,lena); writelen+=lena;}
	if (lene || lenc) createlabel[writelen++]=':';
	if (lene) { memcpy(createlabel+writelen,e,lene); writelen+=lene; }
	if (lenc) { createlabel[writelen++]=':';memcpy(createlabel+writelen,c,lenc); writelen+=lenc; }
	if (otherblockslen) { createlabel[writelen++]=0; createlabel[writelen++]=0; memcpy(createlabel+writelen,otherblocks,otherblockslen); writelen+=otherblockslen;}
	return writelen;
}

LIB_ZMAC_LABEL_NEED_INLINE int zmac_isdeletelabel(char *s)
{
	return (
	s && s[0]=='_' && s[1]==0);
}

LIB_ZMAC_LABEL_NEED_INLINE int zmac_changed(char *a,char *b)
{
	if ((!a && !b) || (a==b)) return 0;
	if (a && b)
	{
		while (*a && *b && *a==*b) {a++;b++;} 
		if (!*a && !*b) return 0;
	}
	return 1;
}

static inline int zmac_changelabel(char *fn,char *reqa,char *reqe,char *reqc)
{
	char readlabel[MAX_STORAGE_LABEL_LENGTH+1],createlabel[MAX_STORAGE_LABEL_LENGTH+1],*otherblocks=NULL,*fa=NULL,*fe=NULL,*fc=NULL,*faold,*feold,*fcold;
	int otherblockslen=0,readlen=0,writelen=0,ret=0;
	if (!(readlen=zmac_getlabel(fn,readlabel))) { fprintf(stderr,"Cannot read ZMAC label from file \"%s\"\n",fn); ret=ZMAC_ERROR_CANNOT_READ_LABEL; goto cleanup; }	// olvasasi hiba
	if (!zmac_extractmultiblock(readlabel,readlen,&fa,&fe,&fc,&otherblocks,&otherblockslen) && !zmac_extractsingleblock(readlabel,&fa,&fe,&fc)) { fprintf(stderr,"Invalid ZMAC label on file \"%s\"\n",fn); ret=ZMAC_ERROR_INVALID_LABEL_ON_FILE; goto cleanup; }
	if (otherblockslen && otherblockslen+(2*MAX_LABEL_LENGTH)+4>MAX_STORAGE_LABEL_LENGTH) { fprintf(stderr,"ZMAC seclabel size exceeded on file \"%s\", cannot modify labels!\n",fn); ret=ZMAC_ERROR_SECLABEL_SIZE_EXCEEDED; goto cleanup; }
	faold=fa; if (zmac_isdeletelabel(reqa)) fa=NULL; else if (reqa && strlen(reqa)) fa=reqa;						// ha toroljuk, akkor toroljuk, ha uresen hagytuk, akkor marad a tartalma, ha beirtunk valamit, akkor arra valtozik
	feold=fe; if (zmac_isdeletelabel(reqe)) fe=NULL; else if (reqe && strlen(reqe)) fe=reqe;
	fcold=fc; if (zmac_isdeletelabel(reqc)) fc=NULL; else if (reqc && strlen(reqc)) fc=reqc;
	if ((!zmac_validlabel(faold) || !zmac_validlabel(feold)|| !zmac_validlabel(fcold)) && (fa || fe|| fc)) { fprintf(stderr,"ZMAC Label on file \"%s\" invalid, cannot modify, only delete!\n",fn); ret=ZMAC_ERROR_SECLABEL_INVALID; goto cleanup; }
	if (zmac_changed(fa,faold) || zmac_changed(fe,feold)|| zmac_changed(fc,fcold))								// ha valtozas van, akkor ki kell irni, kulonben nem
	{
		writelen=zmac_composelabels(createlabel,fa,fe,fc,otherblocks,otherblockslen);
		createlabel[writelen]=0;
		if (!writelen && !zmac_removelabel(fn))				{ fprintf(stderr,"Cannot delete ZMAC label on file: \"%s\".\n",fn); ret=ZMAC_ERROR_CANNOT_DELETE_SECLABEL; goto cleanup; }	// de volt meg az elejen, akkor torolni kell
		if ( writelen && !zmac_setlabel(fn,createlabel,writelen))	{ fprintf(stderr,"Cannot write ZMAC label to file \"%s\"\n",fn); ret=ZMAC_ERROR_CANNOT_WRITE_LABEL; goto cleanup; }
	}
cleanup:
	if (readlen) memset(readlabel,0,readlen);
	if (writelen) memset(createlabel,0,writelen);
	return ret;
}

//
// zmac_return_shell_code
//
// ZeroMAC error codes are negative numbers, but we have to return them as positive to the shell, and 0
// if there are positive ZeroMAC return codes.
//

inline int zmac_return_shell_code(int rc)
{
  return rc<0?-rc:0;
}


